; #########################################################################
;
;                             fl2a18
;
;Procedure for converting a floating point value from the FPU st(0) to ASCII
;Limitation: total number of integer and decimal digits is a maximum of 18
;
;Return: EAX  = 0 if error and null string in memory
;        EAX != 0 if success and ASCII string in memory
;        Rounding is done according to the current RC setting of the FPU
;
;Usage:  invoke fl2a18,Dnum,lpBuf
;
;   where:  
;   Dnum  = immediate value or content of a memory variable
;           for the number of decimal digits required in the ASCII string
;   lpBuf = pointer to a memory buffer for the null-terminated ASCII
;           string (a 24-byte buffer is sufficient)
;
;Example: invoke fl2a18, 15, ADDR szFloat1
;
;Note: The top register is also POPed from the FPU in all cases
;      and the float value destroyed
;
;                       by Raymond Filiatreault
;                              May 2004
;                            MASM  syntax
;
; #########################################################################
;
;      .386                   ; minimum processor needed for 32 bit
;      .model flat, stdcall   ; FLAT memory model & STDCALL calling
;      option casemap :none   ; set code to case sensitive
;
; #########################################################################

.code

fl2a18 proc USES ebx edi Dnum:DWORD, lpBuf:DWORD

LOCAL power9      :DWORD
LOCAL float0      :DWORD
LOCAL bigint      :QWORD
LOCAL highdw      :DWORD
LOCAL lowdw       :DWORD
LOCAL cword       :WORD

      .if Dnum > 18      ;more than 18 decimal places would exceed the limit

   exiterrorp:
            fstp  st
   exiterror:
            xor   eax,eax
            mov   edi,lpBuf
            mov   [edi],al
            ret
      .endif

;check if the content of st(0) is a valid floating point value

      fclex             ;clear exception flags
      fxam
      fstsw ax          ;get results of examination
      fwait
      sahf              ;transfer result to CPU flag register
      .if   ZERO?       ;would be either 0, empty register or denormalized
            jc exiterrorp ;if empty register

      ;A denormalized number would be equivalent to 0 in this context
      ;and considered as valid

            fstp  float0      ;no need for further processing if = 0
   zeroascii:
            mov   edi,lpBuf
            mov   eax,"0"
            mov   [edi],eax
            ret

      .else
            jpo exiterrorp ;would be either unsupported or a NAN
            jc  exiterrorp ;would be infinity
      .endif
      fst   float0      ;copy to memory as REAL4 to retrieve sign later

      mov   power9,1000000000
      mov   ecx,Dnum
      mov   eax,1

;get 10^(Dnum)

      .if   ecx >= 9
            fimul power9
            sub   ecx,9
      .endif

   @@:
      sub   ecx,1
      jc    @F
      imul  eax,10
      jmp   @B

   @@:
      cmp   eax,1
      jz    @F
      mov   lowdw,eax
      fimul lowdw       ;multiply the float by 10^(Dnum)
   @@:
      fabs              ;insure it is positive
      fistp bigint      ;retrieve the boosted float as a 64-bit integer
      fstsw ax          ;get exception flags
      fwait
      shr   al,1
      jc    exiterror   ;invalid operation with too large a number

      fild  bigint      ;load back the 64-bit integer
      fidiv power9      ;get the value of the 10^9 digits
      fstcw cword       ;get control word
      fwait
      mov   ax,cword
      or    ah,0ch      ;change rounding bits to truncate
      push  eax
      fldcw [esp]       ;load modified control word
      fist  highdw      ;store the truncated integer value >10^9 digits
      fldcw cword       ;restore the former control word
      pop   eax         ;clean the stack
      fstsw ax          ;get exception flags
      fwait
      shr   eax,1
      jc    exiterrorp  ;invalid operation with too large a number
      fisub highdw      ;keep only the fractional portion
      fimul power9      ;restore the integer value of the lower 9 digits
      fistp lowdw

      mov   eax,lowdw   ;lower 9 digits now in EAX

      .if   eax == 0 && highdw == 0
            jmp   zeroascii  ;simply display a 0
      .endif

      .if   eax == 1000000000 ;due to rounding of 9's
            xor   eax,eax
            inc   highdw
      .endif

      mov   edi,Dnum    ;use as counter for number of decimal digits
      mov   ebx,9       ;counter for lower 9 digits
      mov   ecx,10
      pushd 0           ;for later use as terminating 0
   @@:
      xor   edx,edx     ;clear for division
      div   ecx         ;remainder (less than 10) -> EDX (effectively DL)
      add   dl,"0"      ;convert to ascii
      push  edx         ;save each digit on stack
      dec   edi         ;bump down count of decimal digits
      .if   ZERO?
            mov   dl,"."
            push  edx   ;insert decimal point
      .endif
      dec   ebx
      jz    @F
      .if   eax != 0 || highdw != 0
            jmp   @B    ;continue conversion
      .endif
      jmp   fl2a18x

   @@:
      mov   eax,highdw
      or    eax,eax
      jz    fl2a18x
   @@:
      xor   edx,edx     ;clear for division
      div   ecx         ;remainder (less than 10) -> EDX (effectively DL)
      add   dl,"0"      ;convert to ascii
      push  edx         ;save each digit on stack
      dec   edi         ;bump down count of decimal digits
      .if   ZERO?
            mov   dl,"."
            push  edx   ;insert decimal point
      .endif
      .if   eax != 0
            jmp   @B    ;continue conversion
      .endif

fl2a18x:
      or    edi,edi     ;would be negative if integer digits already posted
      js    @F
      mov   dl,"0"
      push  edx         ;insert necessary 0s
      dec   edi         ;still used as counter as above
      .if   ZERO?
            mov   dl,"."
            push  edx
      .endif
      jmp   fl2a18x

   @@:
      mov   edi,lpBuf
      mov   eax,float0  ;get the copy of the float
      or    eax,eax     ;check for its sign
      jns   @F
      mov   al,"-"
      stosb
   @@:
      pop   eax         ;retrieve each ascii character from the stack
      stosb
      or    al,al
      jnz   @B          ;continue until terminating 0 is retrieved
      or    al,1
      ret

fl2a18 endp

;end
